//***************************************************************************
// Autor: Jasso Petr (jas063)
// Rok: 2005
//
// Program odesila UDP datagramy na zadanou adresu a port
//
// Povinny argument programu jsou jmeno nebo adresa druheho pocitace a cislo portu
// 
//***************************************************************************


#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <signal.h>
#include <math.h>
#include <sys/times.h>
#include <time.h>
#include <errno.h>

#include "lib.h"

// Zdrojovy port
#define SPORT 1024

// Delitel k vypoctu velikosti packetu podle rychlosti
#define SIZE_DIVIDER 32

// Socket
int sock;

//***************************************************************************
// Funkce vypise napovedu a ukonci program
void help(char *name) {
  printf(
"\n"
"  UDP packet sender. Sends UDP packet of desired size to target host and port every second\n"
"\n"
"  Usage: [-hf] [-s packet_size] host-name port-number\n"
"\n"
"    -h  this help\n"
"\n"
"    -f  flood send\n"
"\n"
"    -s packet_size  sets size of sended packet to packet_size bytes\n"
"\n"
"    -S send_speed   send packets by specified speed (using this option will automatically sets packet size)\n"
"\n", name);

  exit(0);
}

//***************************************************************************
// Funkce nacita parametry
int get_param(int argn, char **arg, int *i, int min) {
  if((*i + 2) > argn) {
    printf("Numeric value expected after parameter %s\n", arg[*i]);
    exit(1);
  }
  int value = atoi(arg[++(*i)]);
  if(value < min) {
    printf("Value must be greater equal than %d!\n", min);
    exit(1);
  }
  
  return value;
}

void clean(void) {
  printf("Try to close socket ...\n");
  close(sock);
  printf("Socket closed.\n");
}
    
void catch_sig(int sig) {
  exit(1);
}
      
timespec get_time() {
  timespec tps;
  
  int ret = clock_gettime(CLOCK_REALTIME, &tps);
  if(ret <  0) {
    perror("Cannot get time: ");
    exit(1);
  }

  return tps;
}

//***************************************************************************
// Hlavni program
int main(int argn, char **arg) {

  if(argn <= 1) help(*arg);

  bool flood = false;
  int port = 0;
  char *host = NULL;

  // Velikost packetu k odeslani
  unsigned packet_size = 60;
  // Rychlost odesilani packetu
  unsigned speed = 0;
  
  
  // Nacteni parametru (host, port)
  for(int i = 1; i < argn; i++) {

    if(!strcmp(arg[i], "-h"))
      help(*arg);

    if(!strcmp(arg[i], "-f"))
      flood = true;

    if(!strcmp(arg[i], "-s")) {
      packet_size = get_param(argn, arg, &i, 60);
      continue;
    }
        
    if(!strcmp(arg[i], "-S")) {
      speed = get_param(argn, arg, &i, 1);
      continue;
    }

    if(*arg[i] != '-') {
      if(!host)
        host = arg[i];
      else if(!port)
             port = atoi(arg[i]);
    }
  }  //for

  if(!host || !port) {
    printf("Invalid or missing host name or port number!\n");
    help(*arg);
  }

  printf("Try to send UDP packets:\n");
  printf("Press CTRL+C for exit ...");

  // Vytvorim IPv4 socket
  int sock = socket(AF_INET, SOCK_DGRAM, 0);
  if(sock == -1) {
    printf("Unable to create a socket!\n");
    exit(1);
  }

  // Vytvorim IPv4 adresu - (host, port)
  in_addr addr_any = { INADDR_ANY };

  sockaddr_in my_addr;
  my_addr.sin_family = AF_INET;	//IPv4
  my_addr.sin_port = htons(SPORT);	//port
  my_addr.sin_addr = addr_any;

  // Zjistim strukturu 'hostent' o druhe strane
  hostent *hostip = gethostbyname(host);
  if(!hostip) {
    printf("Bad host name or IP!\n");
    exit(1);
  }

  sockaddr_in target_addr;
  target_addr.sin_family = AF_INET;	//IPv4
  target_addr.sin_port = htons(port);	//port
  target_addr.sin_addr = *(in_addr*)hostip->h_addr_list[0]; //IP adresa

  // Informace o vytvorenem socketu - my info
  uint lsa = sizeof(my_addr);
  getsockname(sock, (sockaddr*)&my_addr, &lsa);

  // Informace o spojeni
  getpeername(sock, (sockaddr*)&target_addr, &lsa);
  printf("\n\n\tSource port: %d\n\tDestination: %s:%d\n\t", ntohs(my_addr.sin_port), inet_ntoa(target_addr.sin_addr), 
    ntohs(target_addr.sin_port));
  // Pokud jsem zadal rychlost a velikost packetu je vetsi(mensi) nez maximalni(minimalni), upravim je podle rychlosti
  if(speed > 0) {
    packet_size = ((speed * KBPS_BPS) / 8) / SIZE_DIVIDER;

    if(packet_size < MAX_HEADERS_LENGTH)
      packet_size = MAX_HEADERS_LENGTH;

    #ifdef DONT_FRAGMENT
      if(packet_size > MTU)
        packet_size = MTU;
    #endif
    
    printf("Packet size was set to: %d bytes", packet_size);
    printf("\n\tSpeed: %d kbps\n\n", speed);
  }  // if(speed > 0)
  else
    printf("Packet size: %d bytes\n\n", packet_size);
    
  // Zachytavam <CTRL-C>
  signal(SIGINT, catch_sig);
  // Zachytavam SIG_PIPE
  signal(SIGPIPE, catch_sig);
    
  atexit(clean);

  // Vytvorim data o zadane velikosti
  char *data;
  unsigned long data_size = packet_size - HEADERS_SIZE;
  if((data = (char*)malloc(data_size * sizeof(char))) == NULL) { 
    printf("Cannot alocate %d bytes of memory!");
    exit(1);
  }
  memset(data, 0xFF, data_size);
    
  unsigned long sleep;
  if(speed > 0)
    // Vypocet pauz mezi packety	   |- pocet packetu za sekundu                                      -|
    sleep = (unsigned long)round(1000000 / ( ( ((double)(speed * KBPS_BPS)) / 8) / (data_size + HEADERS_SIZE) ) ) / 2;
  else
    // Pokud jsem nezadal rychlost, odesilam co sekundu
    sleep = 1000000;
    
  printf("Sleep = %ld\n", sleep);
  printf("Press CTRL+C for exit ...\n");
  
  timespec before, after;
  // Opakuji
  while(1) {
    before = get_time();
    if(sendto(sock, data, data_size, 0, (struct sockaddr*)&target_addr, sizeof(target_addr)) < 0) {
      printf("Cannot send packet!\n");
      exit(1);
    }
    //usleep(20000);
    after = get_time();
    // Doba odesilani packetu v mikrosekundach
    unsigned long send_time = (after.tv_sec - before.tv_sec) * 1000000 + 
      (after.tv_nsec - before.tv_nsec) / 1000;

    if(!flood)
      usleep(sleep - send_time);	// Cekej sleep
  }
  free(data);
  close(sock);

  return 0;
}
